/*****************************************************************************
    dll.c

    Diese Datei ist Teil des Programmpakets 'ACL-tftp'.
    Alle Rechte bei Andreas G. Schindler
                    Klappacher Str. 2i
                    64285 Darmstadt, Deutschland.

    Dieses Programm ist 'Freie Software'; Sie knnen es entsprechend den 
    Bedingungen der 'Allgemeinen ffentlichen GNU-Lizenz' (verffentlicht 
    durch die 'Free Software Foundation') weiter verbreiten und/oder 
    modifizieren. Massgeblich ist die Version 2 oder eine neuere Version
    dieser Lizenz. 

    Dieses Programm wird in der Hoffnung dass es ntzlich is weitergegeben,
    jedoch OHNE JEDE GARANTIE. Dieser Garantieauschluss bezieht sich auch
    auf Eigenschaften wie Marktfhigkeit oder Eignung zu einem bestimmten
    Zweck. Details finden sich in der 'Allgemeinen ffentlichen GNU-Lizenz'.

    Eine Kopie dieser Lizenz sollte in  Programmpaket, dem diese Datei
    entstammt, enthalten sein. Falls nicht, schreiben Sie an die 
    'Free Software Foundation, Inc.', 59 Temple Place, Suite 330, Boston, 
    MA  02111-1307  USA.


    This file is part of the 'ACL-tftp' Package
    Copyright (C) 1999-2005 by Andreas G. Schindler
                               Klappacher Str. 2i
                               Darmstadt, Germany.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*****************************************************************************/
#include "auto.h"
/*
	This file contains all functions to implement double linkes lists
	for use in tftp access control list processing

	Compile this module with -DTEST to produce a self test program
*/ 

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>

#include "params.h"
#include "dll.h"
#include "dllapps.h"
#include "access.h"


/*
	set up the head/tail pointers for a virgin DLL chain
	the elements of the chain are assumed to be removed
	prior to calling this procedure
*/

void clearChain (dll_t *head)
{
    dll_t *tail  = head+1;

    head->next    = tail->next = tail;
    head->prev    = tail->prev = head;
    head->data[0] = 0;
}

/*
	delete all elements of a DLL chain
	the elements are assumed to be dynamically allocated
	and are thus 'free'ed.
*/

void deleteChain (dll_t *head)
{
    dll_t *ptr, *next, *tail = head+1;

    for (ptr = head->next; ptr != tail; ptr = next) {
        next = ptr->next;
        Free(ptr);
    }

    clearChain(head);
}

/*
	initialize a DLL chain head.
	dynamically allocates the head/tail structs if the
	input argument is set to NULL.
*/

dll_t *initChain(dll_t *head)
{
    if (head == NULL)
        head = (dll_t *)Malloc(2*sizeof(dll_t));

    if (head != NULL)
        clearChain(head);

    return head;
}

/*
	scan backwards from current element to find the chain head
*/

static dll_t *getHead(dll_t *here)
{
    for (;here->prev != here; here = here->prev)
        if (here == NULL)
            break;
    return here;
}

/*
	checks a DLL chain and returns TRUE
	if the chain is empty
*/

int isEmptyChain(dll_t *head)
{
    dll_t *tail = head+1;

    if (head == NULL)
        return -1;

    return ((head->next == tail));
}

/*
	removes an element from a DLL chain.
	controlled by the 2nd (flag) argument, returns
	a pointer to the next (DLL_GETNEXT) or previous
	(DLL_GETPREV) element. DLL_GETPREV is the default,
	as with mostly used forward list scans, this is
	the correct mode. If DLL_FREE is set too, the element
	to be removed is considered object to Free().
*/

dll_t *removeElement (dll_t *elem, int flag, dll_t *head)
{
    dll_t *rp;

    if (elem == NULL)
        return NULL;

    if (head == NULL)
        head = getHead(elem);

    head->data[0]--;

    if ((flag & DLL_GETNEXT))
        rp = elem->next;
    else
        rp = elem->prev;
    elem->prev->next = elem->next;
    elem->next->prev = elem->prev;
    if ((flag & DLL_FREE))
        Free(elem);
    else
        elem->next = elem->prev = NULL;
    return rp;
}

/*
	removes all elements from a chain, that
	have the given 'mark' value. 
	these elements are 'free'ed.
*/

void deleteByMark (dll_t *head, int mark, int flag)
{
    dll_t *ptr, *tail = head+1;

    if (head == NULL)
        return;

    for (ptr = head->next; ptr != tail; ) {
        if (ptr->mark == mark)
            ptr = removeElement(ptr, DLL_GETNEXT|(flag & DLL_FREE), head);
        else
            ptr = ptr->next;
    }
}

/*
	inserts an element at the head (top) of the DLL chain
*/

void insertAtHead (dll_t *head, dll_t *elem)
{
    if (head == NULL || elem == NULL)
        return;

    elem->next = head->next;
    elem->prev = head;
    head->next = elem->next->prev = elem;
    head->data[0]++;
}

/*
	inserts an element at the tail (bottom) of the DLL chain
*/

void insertAtTail (dll_t *head, dll_t *elem)
{
    dll_t *tail;

    if (head == NULL || elem == NULL)
        return;

    tail = head+1;

    elem->prev = tail->prev;
    elem->next = tail;
    tail->prev = elem->prev->next = elem;
    head->data[0]++;
}

/*
	inserts a new element after a particular element of the DLL chain
*/

void insertAfter (dll_t *where, dll_t *elem)
{
    if (where == NULL || elem == NULL)
        return;

    elem->prev  = where;
    elem->next  = where->next;
    where->next = elem->next->prev = elem;
    (getHead(where))->data[0]++;
}

/*
	inserts a new element before a particular element of the DLL chain
*/

void insertBefore (dll_t *where, dll_t *elem)
{
    if (where == NULL || elem == NULL)
        return;

    elem->next  = where;
    elem->prev  = where->prev;
    where->prev = elem->prev->next = elem;
    (getHead(where))->data[0]++;
}

/*
	exchanges the position of 2 elements in a DLL chain
*/

void exchangeElements (dll_t *e1, dll_t *e2)
{
    dll_t *p1, *p2, *n1, *n2;

    if (e1 == NULL || e2 == NULL)
        return;
/*
	save the pointers to the forward- and backward link
*/
    n1 = e1->next;
    p1 = e1->prev;
    n2 = e2->next;
    p2 = e2->prev;
/*
	check if both elements are adjacent to handle this special case
*/
    if (n1 == e2) {		/* consecutive elements e1 -> e2 */
        p1->next = e2;
        n2->prev = e1;
        e1->prev = e2;
        e2->next = e1;
        e1->next = n2;
        e2->prev = p1;
    }
    else if (p1 == e2) {	/* consecutive elements e2 -> e1 */
        p2->next = e1;
        n1->prev = e2;
        e2->prev = e1;
        e1->next = e2;
        e2->next = n1;
        e1->prev = p2;
    }
    else {			/* elements not adjacent */
        e1->next = n2;
        e1->prev = p2;
        e2->next = n1;
        e2->prev = p1;
        n1->prev = p1->next = e2;
        n2->prev = p2->next = e1;
    }
}

/*
	sorts the elements of a DLL chain by the value of
	the 'mark'. the  'flag' argument controls ascending
	(DLL_SORTASC, default) or descending (DLL_SORTDESC)
	mode of sort. sorting arlgorithm is 'sort by exchanging'.
*/

void sortByMark (dll_t *head, int flag)
{
    dll_t *ptr, *next, *tail = head+1;
    int ex;

    if (head == NULL)
        return;

    do {
        ex = 0;
        for (ptr = head->next; ptr != tail; ptr = next) {
            if ((next = ptr->next) == tail)
                break;
            if (flag == DLL_SORTDESC) {
                if (ptr->mark >= next->mark)
                    continue;
            }
            else {
                if (ptr->mark <= next->mark) 
                    continue;
            }
            ex++;
            exchangeElements (ptr, next);
        }
    } while (ex);
}

/*
	get next element of a DLL chain that matches the 'mark'
	starting after element 'start'. with start == NULL, scan
	begins at head of list. no checks are made weather 'start'
	is a valid element of the list specified by 'head'
*/

dll_t *getByMark (dll_t *head, dll_t *start, int mark)
{
    dll_t *ptr, *tail = head+1;

    if (head == NULL)
        return NULL;

    if (start == NULL)
        start = head;

    for (ptr = start->next; ptr != tail; ptr = ptr->next)
        if (ptr->mark == mark)
            return ptr;

    return NULL;
}

/*
	get next element of a DLL chain starting after element 'start'. 
	with start == NULL, scan begins at head of list. 
	no checks are made weather 'start' is a valid element of 
	the list specified by 'head'
*/

dll_t *getNext (dll_t *head, dll_t *start)
{
    dll_t *tail = head+1;

    if (head == NULL)
        return NULL;

    if (start == NULL)
        start = head;

    start = start->next;

    return start == tail ? NULL : start;
}

/*
	get previous element of a DLL chain starting before element 'start'. 
	with start == NULL, scan begins at tail of list. 
	no checks are made weather 'start' is a valid element of 
	the list specified by 'head'
*/

dll_t *getPrev (dll_t *head, dll_t *start)
{
    dll_t *tail = head+1;

    if (head == NULL)
        return NULL;

    if (start == NULL)
        start = tail;

    start = start->prev;

    return start == head ? NULL : start;
}

/*
	adds an elment to the end of a DLL chain. the pattern element
	must be given by an void pointer and its size.
	new memory space is allocated for the new element and
	the values of the pattern element are copied over into
	the new one. the pattern element ist left unchanged.

	attention must be paid, if the pattern element contains
	self-relative pointers, which may produce wrong references
	if copied over to the newly allocated element.
*/

int addPlain (dll_t *head, int mark, void *data, size_t length)
{
    dll_t *new;

    if (head == NULL)
        return -1;

    if ((new = (dll_t *)Malloc(DLL_HEADSIZE + length)) == NULL)
        return -1;

    if ((data != NULL) && length)
        memcpy(new->data, data, length);
    else
        memset(new->data,0,length);

    new->ctype = CTYPE_PLAIN;
    new->mark  = mark;

    insertAtTail (head, new);

    return 0;
}

/*
	check DLL chain for occurence of a loop condition
	forward and backward reference loop checks are controlled
	by the 'flag' values DLL_CHECKPREV (forward loop check) and
	DLL_CHECKNEXT (backward reference loop check).
	only one type of check is performed per call to loopCheck.
*/

dll_t *loopCheck(dll_t *head, int flag)
{
    dll_t *d1, *d2;

    if ((flag & DLL_CHECKNEXT)) {	/* check 'next' chain for backward loops */
        for (d1 = NULL; (d1 = getNext(head, d1)) != NULL;) {
            if (d1->next == d1)
                return d1;
            for (d2 = NULL; (d2 = getNext(head, d2)) != d1;) {
                if (d1->next == d2)
                    return d1;
            }
        }
    }
    else if ((flag & DLL_CHECKPREV)) {	/* check 'prev' chain for foreward loops */
        for (d1 = NULL; (d1 = getPrev(head, d1)) != NULL;) {
            if (d1->prev == d1)
                return d1;
            for (d2 = NULL; (d2 = getPrev(head, d2)) != d1;) {
                if (d1->prev == d2)
                    return d1;
            }
        }
    }
    return NULL;
}


static void dumpout (FILE *fd, dll_t *dp, int i, int atstart)
{
    fprintf(fd,"%c%4d: %-10p  Next: %-10p  Prev: %-10p   Type: %7d  Mark: %d\n",
                    atstart ? '*' : ' ', i, dp, 
                    dp->next, dp->prev, dp->ctype, dp->mark); 
}

/*
	dumps the contents of a DLL chain to 'stderr'
	if 'flag' is DLL_SHOWHEAD, the linkage element's values
	are included in the dump. 'dumpbody' is the pointer to a void
	function that performs the dump of the 'data' (body-) section of
	a chain element. this function is called with 2 arguments:
	1.) a pointer to the current DLL element,
	2.) a pointer to the body section of the current element.
	if flag is 'ored' with DLL_LOOPCHECK, the function 'loopCheck'
	is called in forward and reverse mode and the dump is performed
	on only these elements that are located before the 1st reverse
	loop and after the last forward loop.
*/

void dumpChain(FILE *fd, dll_t *head, dll_t *start, int flag, 
               int bodyflag, void (* dumpbody)())
{

    dll_t *tail, *dp;
    int i, element;
    
    if (fd == NULL)
        fd = stderr;

    tail = head + 1;

    if ((flag & DLL_SHOWHEAD)) {
        if (head == NULL)
            fprintf(fd, "Head :%-10p\n",head);
        else
            fprintf(fd, "Head :%-10p  Next:%-10p  Prev:%-10p  Count:%ld\n",
                    head, head->next, head->prev, head->data[0]);

        if (tail == NULL)
            fprintf(fd, "Tail :%-10p\n",tail);
        else
            fprintf(fd, "Tail :%-10p  Next:%-10p  Prev:%-10p\n",
                    tail, tail->next, tail->prev);

        if (start == NULL)
            fprintf(fd, "Start: %-10p\n",start);
        else
            fprintf(fd, "Start:%-10p  Next:%-10p  Prev:%-10p\n",
                    start, start->next, start->prev);
    }

    if (head == NULL) return;

    fprintf (fd, "Starting loop count check\n");
    for (dp = NULL, i=0; (dp = getNext(head,dp)) != NULL; i++);
    fprintf (fd, "Head counter: %d  Elements: %d  %s\n", 
             (int)head->data[0], i,
             (i == (int)head->data[0]) ? "o.k." : "*** Mismatch ***");

 
    if ((flag & DLL_LOOPCHECK)) {
        dll_t *db, *df;

        db = loopCheck(head, DLL_CHECKNEXT);
        df = loopCheck(head, DLL_CHECKPREV);

        fprintf (fd, "Starting backward loop check\n"); 
        if (db != NULL) {
            if (df != NULL) {
                i = 0;
                dp = NULL;
                element = 0;
                do {
                    dp = getNext(head, dp);
                    if ((flag & DLL_SHOWHEAD))
                        dumpout(fd, dp, ++i, (dp == start));
                    if (dumpbody != NULL)
                        (dumpbody)(fd, dp, bodyflag, ++element);
                } while (dp != db);
            }
            fprintf (fd, "Backward loop from %10p to %10p\n", db, db->next); 
        }
        fprintf (fd, "Backward loop check done\n"); 
        fprintf (fd, "Starting forward loop check\n"); 
        if (df != NULL) {
            if (db != NULL) {
                i = 0;
                dp = NULL;
                element = 0;
                do {
                    dp = getPrev(head, dp);
                    if ((flag & DLL_SHOWHEAD))
                        dumpout(fd, dp, --i, (dp == start));
                    if (dumpbody != NULL)
                        (dumpbody)(fd, dp, bodyflag, ++element);
                } while (dp != df);
            }
            fprintf (fd, "Forward loop from %10p to %10p\n", df, df->prev); 
        }
        fprintf (fd, "Forward loop check done\n"); 
        if ((df != NULL) | (db != NULL))
            return;
    }

    i = 0;
    dp = NULL;
    element = 0;
    while ((dp = getNext(head, dp)) != NULL) {
/*
        if ((flag & DLL_SHOWHEAD))
            dumpout(fd, dp, ++i, (dp == start));
*/
        if (dumpbody != NULL)
            (dumpbody)(fd, dp, bodyflag,++element);
    }
}
        
dll_t *findByMark (dll_t *head, int mark, int how)
{
    dll_t *dp, *df;

    df = (how = DLL_GETFIRST) ? NULL : head->next;

    for (dp = NULL; (dp = getNext(head, dp)) != NULL;) {
        switch (how) {
          case DLL_GETMAX:
            if (dp->mark > df->mark)
                df = dp;
            break;
          case DLL_GETMIN:
            if (dp->mark < df->mark)
                df = dp;
            break;
          case DLL_GETFIRST:
            if (dp->mark == mark)
                return dp;
            break;
          case DLL_GETLAST:
            if (dp->mark == mark)
                df = dp;
            break;
        }
    }
    return df;
}


#ifdef TEST

#include "time.h"
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include "util.h"
#include "dllapps.h"

#define DumpChain dumpChain(NULL,dp,NULL,DLL_SHOWHEAD|DLL_LOOPCHECK,(dp == NULL),dumpBody)

static char *regcerror(int err)
{
    static char errbuf[128];

    if (!regerror(err, NULL, errbuf, sizeof(errbuf)))
        return "Non REGEX error";
    else
        return errbuf;
}



int main (int argc, char **argv)
{
    dll_t *dp, *dx, dlist[6];
    static char *relist[]= { "^file1\\....", "file[1-5]\\.dat", "[a-z].*\\.dat",
                              ".*file1.dat", "file.\\.dat", "nope" };
    static char *mplist[]= { "%b.%E",        "%B.%H%M%S",   "%Dsub2/%B", 
                             NULL,           NULL,          NULL };
    static char *cklist[] = { "file1.xxx", "file5.dat", "/any/other.dat", 
                              "/myfile1.dat", "file0.dat", "nope" };
    static char *alist[] = { "1.2.3.4",         "1.2.3.9"  , "8.8.8.8", "1.2.9.9",     "2.2.2.2" };
    static char *mlist[] = { "255.255.255.255", "0.0.0.255", "0.0.0.0", "0.0.255.255", "0.0.0.0" };
    static char *xlist[] = { "1.2.3.4",         "1.2.3.5",   "8.8.8.8", "1.2.4.5",     "9.9.9.9" };
   
    int i, rtc;
    struct in_addr ia, im;

    puts ("double linked list test program");

    puts ("Step 1: creating a list");

    if ((dp = initChain(NULL)) == NULL) {
        puts("Cannot create Chain");
        exit(1);
    }
    DumpChain;

    for (i=0; i<(sizeof(dlist)/sizeof(dll_t)); i++) {
        dlist[i].mark = 1 + i/3;
        dlist[i].data[0] = i;
        dlist[i].next = dlist[i].prev = NULL;
    }

    puts ("Step 2: inserting 2 elements at head of the list");

    insertAtHead(dp, &dlist[1]);
    insertAtHead(dp, &dlist[0]);
    DumpChain;

    puts ("Step 3: inserting 2 elements at tail of the list");

    insertAtTail(dp, &dlist[4]);
    insertAtTail(dp, &dlist[5]);
    DumpChain;

    puts ("Step 4: inserting 2 elements in the middle of the list");

    insertAfter(&dlist[1], &dlist[2]);
    insertBefore(&dlist[4], &dlist[3]);
    DumpChain;

    puts ("Step 5: exchanging elements");


    puts ("elements 0/1 (consecutive ascending)");
    exchangeElements(&dlist[0], &dlist[1]);
    DumpChain;
    puts ("elements 0/1 (consecutive descending)");
    exchangeElements(&dlist[0], &dlist[1]);
    DumpChain;

    puts ("elements 2/4 (non consecutive)");
    exchangeElements(&dlist[2], &dlist[4]);
    DumpChain;

    puts ("elements 3/5 (non consecutive at end of list)");
    exchangeElements(&dlist[3], &dlist[5]);
    DumpChain;

    puts ("Step 6: selecting by mark");

    for (dx = dp; dx != NULL;) {
        dx = getByMark(dp, dx, 2);
        if (dx != NULL) printf ("Selection 2 retrieves %p Data: %d\n", 
                                dx,(int)dx->data[0]);
    }

    puts ("Step 7: sorting by mark");

    sortByMark (dp, DLL_SORTASC);
    DumpChain;

    puts ("Step 8: deleting 3 elements (1st, 3rd, last)");

    removeElement(&dlist[0], DLL_GETPREV, dlist);
    removeElement(&dlist[2], DLL_GETPREV, dlist);
    removeElement(&dlist[5], DLL_GETPREV, NULL);	/* test getHead() */
    DumpChain;

    puts ("Step 9: deleting the rest of the list");

    clearChain(dp);
    DumpChain;

    puts ("Step 10: setting up a regex list");

    for (i=0; i<(sizeof(relist)/sizeof(char *)); i++)
        if ((rtc = addRePattern ((dlr_t *)dp,i,relist[i],mplist[i],0,127,0,NULL))) {
            char *msgtxt;
            if (rtc == -1)
                msgtxt = "no dynamic memory";
            else
                msgtxt = regcerror(rtc);
            printf("Cannot insert regex %d [%s]: %d:%s\n",
                   i,relist[i],rtc,msgtxt);
            exit(1);
        }
    DumpChain;

    puts ("Step 11: looking up by regex match");

    for (i=0; i<(sizeof(cklist)/sizeof(char *)); i++) {
        dlr_t *dr = NULL;
        int c = 0;
        while ((dr = findRePattern((dlr_t *)dp, dr, 0, cklist[i], 1)) != NULL)
            printf("Looking up [%s] returns: %d [%s] (%s)->(%s)\n",
                   cklist[i],++c,relist[dr->mark], 
                   (mplist[dr->mark] == NULL) ? "" : mplist[dr->mark],
                   (mplist[dr->mark] == NULL) ? "" : 
                   xlateFileMap(cklist[i], mplist[dr->mark], NULL, NULL, 1));
        if (!c)
            printf("Looking up [%s] returns: *not found*\n", cklist[i]);
    }

    puts ("Step 12: deleting list element by mark (3)");

    DumpChain;
    deleteByMark (dp, 3, 0);
    DumpChain;
    clearChain  (dp);

    puts ("Step 13: setting up an ip addr/mask list");

    for (i=0; i<(sizeof(alist)/sizeof(char *)); i++)
        if (addACL ((dla_t *)dp, 1, i, alist[i], mlist[i], 0, NULL)) {
            printf("Cannot insert ACL %d [%s/%s]\n",i,alist[i],mlist[i]);
            exit(1);
        }
    DumpChain;

    puts ("Step 14: looking up by ip address");

    for (i=0; i<(sizeof(xlist)/sizeof(char *)); i++) {
        dla_t *da = NULL;
        int c = 0;
        in_addr_t aa = (in_addr_t)inet_addr(xlist[i]);

        while ((da = findACL((dla_t *)dp, da, 1, aa, FIND_MAX)) != NULL) {
            ia.s_addr = da->acl.addr;
            im.s_addr = da->acl.mask;
            printf("Looking up [%s] returns: %d [%d]",
                   xlist[i], ++c, da->acl.perm);
            printf("  %-16s", inet_ntoa(ia));
            printf("%-16s\n", inet_ntoa(im));
        }

        if (!c)
            printf("Looking up [%s] returns: *not found*\n", xlist[i]);
    }

    return 0;
}

#endif /* TEST */
